# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

mat = """a Matlab coded irritation design of on- and offsets."""

########################################################################
# General arguments
########################################################################

session = """path to a session file or a template which defines the path
where to find or save the respected session files when using a protocol
file."""

sfit = """path to a result file or a template which defines the path
where to find or save the respected result files when using a protocol
file."""

population_space = """name of a population space."""

population_space_name = """name of a population space."""

population_space_directory = """directory in which to save the
population space."""

population_map = """instance of a population map."""

irritation = """irritation file.  File name containing the irritation
instance for this session.  An irritation instance contains all
information of the paradigm of this session, i.e., whether the paradigm
follows a block design, the names of the respective blocks, their
respective onsets and respective durations."""

template = """template file used for the population space. The file
should contain a 3D-image of a brain in any file format understood by
the NiBabel project, e.g, any of ANALYZE (plain, SPM99, SPM2 and later),
GIFTI, NIfTI1, NIfTI2, MINC1, MINC2, MGH and ECAT as well as Philips
PAR/REC (For more details see http://nipy.org/nibabel/). On the other
hand, it should also be understood by FSL."""

template_mask = """template mask file used for the population space. The
file should contain a 3D-image of a brain in any file format understood
by the NiBabel project, e.g, any of ANALYZE (plain, SPM99, SPM2 and
later), GIFTI, NIfTI1, NIfTI2, MINC1, MINC2, MGH and ECAT as well as
Philips PAR/REC (For more details see http://nipy.org/nibabel/). On the
other hand, it should also be understood by FSL."""

reference_maps = """reference maps."""

########################################################################
# Miscellaneous
########################################################################

verbose = """increase output verbosity."""

quite = """do not be verbose."""

force = """by default {} files will not be overwritten but parsed
instead.  This will force to re-create any of these files."""

conditional_force = """by default {} files will not be overwritten but
parsed instead.  This will force to re-create any of files on which
creation this file depends has a newer time stamp."""

fix = """by default {} files will not be overwritten but parsed
instead.  This will force to re-create any of these files."""

remove = """by default {} files will not be overwritten but parsed
instead.  This will force to re-create any of these files."""

skip = """do not create or fit anything"""

ignore_lock = """You can either ignore the lock, remove the lock, or
respect the lock.  Respect the lock: skip protocol entry. Remove the
lock, removes the lock and skips the protocol entry.


ignore whether a file has been locked by a different
program. Only use this if orphans exists, and you now for sure no other
program is running that may have looked the respected files. Together
with --skip this may be used to delete orphan lock files. Short -is"""

delete_invalid = """delete all {} files which entries in the protocol
which are marked as invalid"""

remove_lock = """remove lock"""

########################################################################
# Arguments specific for using the protocol API
########################################################################

protocol = """
protocol file.
"""

protocol_log = """
Log file. Entries that could not be processed are marked as invalid
"""

cohort = """only use entries instances which belong to this cohort. If
no protocol file is used, then this is the name of the cohort to which
this subject belongs."""

j = """only use entries which belong to this subject id. If no protocol
file is used, then this is the id of the subject in its cohort."""

datetime = """only use entries which are stamped with this date and
time."""

paradigm = """only use entries which belong to this irritation design.
The name of the irritation will appear as part of the default file names
produced by various fmristats command line tools.  Unless you are
planing to provide custom file names to all of fmristats' command line
tools, it is recommend not to use any special characters in this
name."""

strftime = """convert time to string according to this format specification."""

########################################################################
# Arguments specific for the setup of a session instances
########################################################################

nii = """
the output of a fMRI session.  The file should contain a 4D-image of a
fMRI session in any file format understood by the NiBabel project, e.g,
any of ANALYZE (plain, SPM99, SPM2 and later), GIFTI, NIfTI1, NIfTI2,
MINC1, MINC2, MGH and ECAT as well as Philips PAR/REC.  For more details
see http://nipy.org/nibabel/.  Please note that fmristats has only been
tested with Nifti1 files.
"""

epi_code = """
code for the direction of the normal vector of the EPIs assuming the
data is in RAS+. For example, if the EPIs have been measured inferior to
superior, then set EP to 2, if they have been measured superior to
inferior, set EP to -2, if they have been measured posterior to
anterior, i.e., they are parallel to the left--right-inferior--superior
plain, then set EP to 1.  Values in a protocol file will take precedence
"""

onsetx = """times of onsets of condition x, which will be interpreted
as the control irritation of the subject.  Should, e.g., be given in the
form:

                --onsetx 2 84 166 248 330

to be parsed correctly.  If onsets are not given in order, they will be
sorted automatically.  Must be specified if --irritation is not.
"""

onsety = """times of onset of condition y, which will be interpreted as
the stimulus irritation of the subject.  Should, e.g., be given in the
form:

                --onsety 37 119 201 283

to be parsed correctly.  If onsets are not given in order, they will be
sorted automatically.  Must be specified if --irritation is not.  """

durationx = """duration of the irritation of the subject under condition
x.  Must be specified if --irritation is not."""

durationy = """duration of the irritation of the subject under condition
y.  Must be specified if --irritation is not.  """

detect_foreground = """detect foreground."""

set_foreground = """set foreground."""

grubbs = """an outlier detection is performed to identify scans which
may have been acquired during severe head movements. More precisely, a
Grubbs' outlying test will be performed on the set of estimated
principle semi axis for each full scan cycle on the given level of
significance.

When using fmririgid to create ReferenceMaps, the default is 0.1, and
the information of outlying scans is saved to disk together with the
estimated rigid body transformations. Then, when running fmrifit,
this information is used. When setting --grubbs in fmrifit, outlier
estimation is performed again."""

no_tracking = """skip tracking of the rigid body, i.e., skip fitting
head movements to the data.  This will have the effect that head
movements are set to the identity on the reference space.  This is
useful if your data consists of phantom data.  """

########################################################################
# Arguments specific for the RSM Signal Model: design matrix
########################################################################

burn_in = """acquisition burn in. The *first* scan cycle that should
enter the model."""

time_order = """degree of polynomial that models temporal fluctuation.
Default is 1."""

drop_irritation = """if irritation information is present in the model,
here, you may drop it."""

########################################################################
# Arguments specific for the RSM Signal Model
########################################################################

scale = """
standard deviation of a Gaussian kernel that defines the weighting
scheme of the underlying WLS regression.  If not given explicitly, scale
will be set to one half of the length of the diagonal of the
orthorhombic measure lattice of the session (and this is recommended).
This default behaviour corresponds to the SCALE_TYPE ``diagonal`` and
can be overwritten by setting SCALE_TYPE to a different value. The
parameter SCALE will determine the final curvature of the fitted effect
field. The larger SCALE, the flatter the fitted effect field will
appear.
"""

scale_type = """
should be either one of ``diagonal``, ``max``, or ``min``.

If SCALE_TYPE is ``diagonal``, then scale is equal to one half of the
length of the diagonal of the orthorhombic measure lattice.  If
SCALE_TYPE is ``min`` (or ``max``), then SCALE is one half of the
minimal (or maximal) edge length of the orthorhombic measure lattice.
"""

stimulus_block = """
name of the stimulus block.
"""

control_block = """
name of the control block.
"""

population_mask = """Fit the effect field **only** at points which are
within the template brain of the population space.  You should only use
this option, if you are sure that your diffeomorphism from population to
reference space of this sample is dead on. If this is the case -- and it
should be, right? -- this will considerably speed up the computation
time.  It does, however, make visual control of the fit more difficult,
and should be used with care."""

factor = """
Only observations within an area of FACTOR standard deviations under
the Gaussian distribution defined by SCALE will find their way into the
WLS regression.
"""

mass = """
Only observations within an area of mass `mass` under the Gaussian
distribution defined by SCALE will find their way into the WLS
regression.  If specified, must be a value between 0 and 1, and should
be a value close to 1.
"""

offset_beginning = """
The haemodynamic response to stimulus is not immediate.  It is usually
assumed that the HR-function spikes approximately five seconds after the
first stimulus.  The value of OFFSET_BEGINNING is added to the onset
times of the irritation phases to allow you to wait till the subject's
brain is in the respected irritation modus.
"""

offset_end = """
Similar to OFFSET_BEGINNING the value OFFSET_END is removed from the end
of each irritation phase and not considered in the fitting.
"""


########################################################################
# Arguments specific for the RSM Population Model
########################################################################

########################################################################
# Epilog
########################################################################

epilog = """
For more information visit https://fmristats.github.io/
"""

cores = """
Number of cores to use. Default is the number of cores on the machine
"""



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .name import Identifier

import os

import datetime

import pickle

class Lock:
    """
    An instance of this class may by used to temporary look
    the file name while processing or fitting data.

    Parameters
    ----------
    name : Identifier
    who : str
        name of the program that currently locks this file.
    fname :
        file name that is being locked.
    timestamp : datetime
        time of lock.
    """
    def __init__(self, name, who, fname, timestamp=None):
        assert type(name) is Identifier, 'name must be an Identifier'

        if timestamp is None:
            timestamp = datetime.datetime.now()

        self.name = name
        self.fname = fname
        self.who = who
        self.timestamp = timestamp

    def describe(self, strftime='%Y-%m-%d-%H%M'):
        description = """
        Cohort:     {}
        Subject:    {:d}
        Paradigm:   {}
        File name:  {}
        Created on: {}
        Looked by:  {}
        """
        return description.format(
                self.name.cohort,
                self.name.j,
                self.name.paradigm,
                self.fname,
                self.timestamp.strftime(strftime),
                self.who,
                )

    def unlock(self):
        os.remove(self.fname)

    def conditional_unlock(self, df, index, verbose, force=False):
        if (not df.ix[index,'valid']) or force:
            if verbose:
                print('{}: Unlock: {}'.format(self.name.name(), self.fname))
            self.unlock()
            df.ix[index,'locked'] = False
            return True
        else:
            return False

    def save(self, file=None, **kwargs):
        """
        Save instance to disk

        This will save the instance to disk.

        Parameters
        ----------
        file : str
            File name.
        """
        if file is None:
            file = self.fname
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Defines classes for affine transformations, rigid body transformations,
and orthonormal transformations, and functions for manipulating and
calculating inverses of affine transformations.

"""

from .euler import rotation_matrix_to_euler_angles

# TODO: make importing nibabel obsolete, here!
import nibabel as ni

import pickle

import numpy as np

from numpy.linalg import inv, det, norm, svd

########################################################################
#
# Basic functions
#
########################################################################

def cartesian2homogeneous(mat, vec):
    return ni.affines.from_matvec(mat, vec)

def from_cartesian(mat, vec):
    return Affine(cartesian2homogeneous(mat, vec))

def is_rotation_matrix(a, **kwargs):
    """
    Test if a 3x3-matrix is a rotation matrix

    Parameters
    ----------
    a : ndarray, shape (n,n)
        an darray
    **kwargs :
        parameters passed on to `np.close`.

    Returns
    -------
    bool :
        True or False.

    Notes
    -----
    The function assumes that `a` is square.
    """
    return np.isclose(
            np.dot(np.transpose(a), a),
            np.identity(3), **kwargs).all()

def isclose(x, y, **kwargs):
    """
    Test if two affines are close to each other

    Parameters
    ----------
    x : Affine
        an affine transformation.
    y : Affine
        an affine transformation.
    **kwargs :
        parameters passed on to `np.close`.

    Returns
    -------
    bool :
        True or False.
    """
    assert type(x) is Affine and type(y) is Affine, 'x and y must be of type Affine'
    return np.isclose(x.affine, y.affine, **kwargs).all()

def isinverse(x, y, **kwargs):
    """
    Test if two affines are inverses of each other

    Parameters
    ----------
    x : Affine
        an affine transformation.
    y : Affine
        an affine transformation.
    **kwargs :
        parameters passed on to `np.close`.

    Returns
    -------
    bool :
        True or False.
    """
    assert type(x) is Affine and type(y) is Affine, 'x and y must be of type Affine'
    return np.isclose(x.dot(y).affine, np.eye(4), **kwargs).all()


########################################################################
#
# Class for working with a single affine transformation
#
########################################################################

class Affine:
    """
    An affine transformation

    Affine transformations are stored in homologous coordinates.

    Parameters
    ----------
    affine : ndarray, shape (4,4)
        An affine transformation in homologous coordinates.
    """
    def __init__(self, affine):
        assert affine.shape == (4,4), 'wrong shape'
        assert (affine[3] == np.array([0,0,0,1])).all(), '3rd row must be [0,0,0,1]'
        self.is_rigid = is_rotation_matrix(affine[:3,:3])
        self.affine = affine

    def apply(self, x):
        """
        Apply affine transformation

        Will apply the affine transformation to the matrix or vector x.

        Parameters
        ----------
        x : ndarray, shape (3,) or (...,3)
            Vector or matrix of (row) vectors.

        Return
        ------
        ndarray:
            The result of the affine applied to the matrix (or vector)
            x, in particular, it will be of the same shape.
        """
        tmp = np.append(x,np.ones(x.shape[:-1] + (1,)), axis=-1)
        return np.einsum('...ij,...j', self.affine, tmp)[...,:3]

    def apply_to_index(self, index):
        """
        Apply affine transformation to index

        Parameters
        ----------
        index : 3-tuple
            3-tuple of int

        Returns
        -------
        ndarray:
            The result of the affine applied to the index.
        """
        return self.apply(np.array(index))

    def apply_to_indices(self, indices):
        """
        Apply affine transformation to indices

        Parameters
        ----------
        index : 3-tuple
            3-tuple of Slice or Ellipse objects.

        Returns
        -------
        ndarray:
            The result of the affine applied to the indices.
        """
        return self.apply(np.moveaxis(np.mgrid[indices], 0, -1))

    def inv(self):
        """
        Calculates the inverse of an affine transformation

        If the affine transformation is rigid, i.e., of the form A⋅x + b
        with A being a rotation matrix, then a more numerical stable
        algorithm for calculating the inverse is used than otherwise.

        Return
        ------
        The inverse of the affine transformation.
        """
        if self.is_rigid:
            u = self.affine[:3,:3]
            p = self.affine[:3, 3]

            inverse = np.empty_like(self.affine)
            inverse[:3,:3] =  u.T
            inverse[:3, 3] = -u.T.dot(p)
            inverse[3] = (0,0,0,1)

            return Affine(inverse)
        else:
            return Affine(inv(self.affine))

    def dot(self, x):
        """
        Matrix multiplication with x

        Parameters
        ----------
        x : Affine or Affines
            An affine transformation.

        Returns
        -------
        Affines or Affines:
            Output has the same type as x.
        """
        assert (type(x) is Affine) or (type(x) is Affines), 'x must be Affine or Affine'

        if type(x) is Affine:
            return Affine(self.affine.dot(x.affine))

        if type(x) is Affines:
            results = np.empty_like(x.affines)
            for t in range(x.n):
                results[t] = self.affine.dot(x.affines[t])
            return Affines(results)

    def euler(self):
        """
        Calculate Euler angles
        """
        assert self.is_rigid, 'does only work if affine is rigid'
        return rotation_matrix_to_euler_angles(self.affine[:3,:3])

    def resolution(self):
        """
        Resolution

        The resolution is the length of the base vectors (1,0,0),
        (0,1,0), and (0,0,1) after being mapped to by this affine
        transformations. The length of ``A⋅b``, for ``A`` being this
        affine transformation and ``b`` being one of the three basis
        vectors is the norm of ``A⋅b - A⋅0``.
        """
        return norm(self.affine, axis=0)[:3]

    def volume(self):
        """
        Volume of the parallelepiped spanned by the image of the
        canonical basis
        """
        return abs(det(self.affine))

    def diagonal(self):
        return norm(self.resolution())

    def aspect(self, x, y):
        """
        Parameters
        ----------
        x : int, either 0,1,2
            x
        y : int, either 0,1,2
            y

        Returns
        -------
        float :
            The aspect ratio of the respected resolution.
        """
        resolution = self.resolution()
        return resolution[x] / resolution[y]

    def describe(self):
        description = """
        Resolution (left to right):         {:.2f} mm
        Resolution (posterior to anterior): {:.2f} mm
        Resolution (inferior to superior):  {:.2f} mm
        Diagonal:      {:9,.2f}
        Volume:        {:9,.2f}
        Aspect 0 on 1: {:9,.2f}
        Aspect 0 on 2: {:9,.2f}
        Aspect 1 on 2: {:9,.2f}
        Rigid transformation: {}"""
        resolution = self.resolution()
        return description.format(
                resolution[0],
                resolution[1],
                resolution[2],
                self.diagonal(),
                self.volume(),
                self.aspect(0,1),
                self.aspect(0,2),
                self.aspect(1,2),
                self.is_rigid
            )

    def save(self, file, **kwargs):
        """
        Save instance to disk

        Parameters
        ----------
        file : str
            A file name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

########################################################################
#
# Class for working with an array of affine transformations
#
########################################################################

class Affines:
    """
    Affine transformations

    Affine transformations are stored in homologous coordinates.
    """
    def __init__(self, affines):
        """
        Parameters
        ----------
        affines : ndarray, shape (n,4,4)
            Affine transformations in homologous coordinates.
        """
        # TODO: allow for higher dimesional indices: affines.shape =
        # (m,n,4,4)
        n, x, y = affines.shape
        assert (x,y) == (4,4), 'affines must be given in homologous coordinates'
        assert np.isclose(affines[:,3], np.array((0,0,0,1))).all(), 'last row must be [0,0,0,1]'
        rigid = np.repeat(False, n)
        for t in range(n):
            rigid[t] = is_rotation_matrix(affines[t,:3,:3])
        self.rigid = rigid.all()
        self.affines = affines
        self.n = n

    def index(self, j):
        return Affine(self.affines[j])

    def mean(self):
        """
        Calculates the mean affine transformation

        Returns
        -------
        Affine :
            The mean affine transformation.

        Notes
        -----
        Some basic non-Euclidean statistics is used here.
        """
        # https://en.wikipedia.org/wiki/Spherical_coordinate_system
        # https://en.wikipedia.org/wiki/Mean_of_circ
        return Affine(self.affines.mean(axis=0))

    def mean_rigid(self):
        """
        Calculates the mean affine transformation of rigid
        transformations

        Returns
        -------
        Affine :
            The mean rigid affine transformation.

        Notes
        -----
        Some basic non-Euclidean statistics is used here.
        """
        assert self.rigid, 'affines must be rigid'
        m = self.mean().affine
        u, s, v = svd(m[:3,:3])
        m[:3,:3] = u.dot(v)
        return Affine(m)

    def euler(self):
        """
        Calculate Euler angles for all affine transformations

        Notes
        -----
        This only makes sense, if the affine transformations are rigid.
        """
        assert self.rigid, 'affines must be rigid'
        euler = np.zeros((self.n,3))
        for t in range(self.n):
            euler[t] = rotation_matrix_to_euler_angles(self.affines[t,:3,:3])
        return euler

    def apply(self, x):
        """
        Apply affine transformations to a vector

        Will apply the affine transformations to the vector x.

        Parameters
        ----------
        x : ndarray, shape (3,)
            Vector to which the affine transformations are applied

        Returns
        -------
        The result of the affine transformations applied to x.
        """
        assert x.shape == (3,), 'only makes sense if x is a 3-dim vector'
        tmp = np.append(x,np.ones(x.shape[:-1] + (1,)), axis=-1)
        return np.einsum('...ij,...j', self.affines, tmp)[...,:3]

    def apply_to_index(self, index):
        """
        Apply affine transformation to index

        Parameters
        ----------
        index : 3-tuple
            3-tuple of int

        Returns
        -------
        ndarray:
            The result of the affine applied to the index.
        """
        return self.apply(np.array(index))

    def inv(self):
        """
        Calculates the inverse of an affine transformation

        If the affine transformation is rigid, i.e., of the form A⋅x + b
        with A being a rotation matrix, then a more numerical stable
        algorithm for calculating the inverse is used than otherwise.

        Return
        ------
        The inverse of the affine transformation.
        """
        if self.rigid:
            u = self.affines[:,:3,:3]
            p = self.affines[:,:3, 3]
            uT = np.moveaxis(u, [1,2], [2,1])

            inverses = np.empty_like(self.affines)
            for t in range(self.n):
                inverses[t,:3,:3] =  uT[t]
                inverses[t,:3, 3] = -uT[t].dot(p[t])

            inverses[:,3] = (0,0,0,1)
            return Affines(inverses)
        else:
            inverses = np.empty_like(self.affines)
            for t in range(self.n):
                inverses[t] = inv(self.affines[t])
            return Affines(inverses)

    def dot(self, x):
        """
        Matrix multiplication (a.k.a. dot product) with the affines
        transformation x

        This will performe a matrix multiplication of each affine
        transformation with x.

        Parameters
        ----------
        x : Affine
            An affine transformation.

        Returns
        -------
        The result of the calculation is again of type Affines.
        """
        assert type(x) is Affine, 'x must be Affine'
        return Affines(self.affines.dot(x.affine))

    def save(self, file, **kwargs):
        """
        Save instance to disk

        Parameters
        ----------
        file : str
            A file name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Functions that return the image of a coordinate given in one space in
the coordinates of the image space of a given map.

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

    parser.add_argument('input',
        help="""a population space, a population map, any
        diffeomorphism, or an image file.""")

    parser.add_argument('first',  help='first coordinate',  type=float)
    parser.add_argument('second', help='second coordinate', type=float)
    parser.add_argument('third',  help='third coordinate',  type=float)

    parser.add_argument('-i', '--index',
        action='store_true',
        help="""by default coordinates are assumed to live in Euclidean
        space. If --index is set, coordinate are interpreted to live in
        index space instead.""")

    parser.add_argument('--inverse',
        action='store_true',
        help="""inverse""")

    parser.add_argument('-v', '--verbose',
            action='store_true',
            help='verbose output')

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ...load import load

from ...pmap import PopulationMap

from ...diffeomorphisms import Warp, Displacement

import sys

import numpy as np

########################################################################

def call(args):
    """
    Functions that return the image of a coordinate given in one space in
    the coordinates of the image space of a given map.

    Parameters
    ----------
    args : Arguments
    """
    try:
        x = load(args.input)
    except Exception as e:
        print('Cannot read: {}'.format(args.input))
        print('Failed with: {}'.format(e))
        sys.exit()

    if type(x) is PopulationMap:
        x = x.diffeomorphism

    if args.index:
        index = tuple(np.array((
            args.first,
            args.second,
            args.third), dtype=int))

        if args.verbose:
            print(
"""index: {}
has coordinates in the domain: {}
and coordinates in the image:  {}""".format(
                index,
                x.reference.apply_to_index(index),
                x.apply_to_index(index)))
        else:
            print(x.apply_to_index(index))

    else:
        coordinate = np.array((
            args.first,
            args.second,
            args.third), dtype=float)

        if args.inverse:
            if type(x) is Warp or type(x) is Displacement:
                coordinates = x.coordinates()
                distances   = ((coordinates - coordinate)**2).sum(axis=-1)
                domain_index = np.unravel_index(np.argmin(distances), distances.shape)
                domain_coordinate = x.reference.apply_to_index(domain_index)
            if args.verbose:
                print(
"""index: {}
has coordinates in the domain: {}
and coordinates in the image:  {}""".format(
                    domain_index,
                    domain_coordinate,
                    coordinate))
            else:
                print(domain_coordinate)

        else:
            if args.verbose:
                print(
"""index: {}
has coordinates in the domain: {}
and coordinates in the image:  {}""".format(
                    x.reference.inv().apply(coordinate),
                    coordinate,
                    x.apply(coordinate)))
            else:
                print(x.apply(coordinate))



# Copyright 2015-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Irritations, paradigms.

"""

from .name import Identifier

import numpy as np

import pickle

class Irritation:
    """
    Irritation design

    Parameters
    ----------
    name : Identifier
        identifier of the session
    """

    def __init__(self, name):
        assert type(name) is Identifier, 'name must be of type Identifier'
        self.name = name

    def design(self, slice_time):
        return np.zeros_like(slice_time)

    def describe(self):
        """
        Give a description of the instance
        """
        return self.name.describe()

    def save(self, file, **kwargs):
        """
        Save irritation instance to disk

        This will save the irritation instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

class Block(Irritation):
    """
    Block irritation of the subject

    Parameters
    ----------
    name : Identifier
        identifier of the session
    names : list
    onsets : dict
    durations : dict
    """

    def __init__(self, name, names, onsets, durations):
        assert type(name) is Identifier, 'name must be of type Identifier'

        number = len(names)

        assert len(names) == number, 'number of names does not match number of blocks'
        assert len(durations) == number, 'number of durations does not match number of blocks'
        assert len(onsets) == number, 'number of onsets does not match number of blocks'

        self.name = name
        self.number = number
        self.names = names
        self.durations = durations
        self.onsets = onsets

#    def design(self, slice_time, s, c, offset=0, preset=0):
#        """
#        Defines the irritation design vector
#
#        Parameters
#        ----------
#        slice_time : ndarray, shape (n,) or (n,m)
#            times at which slices have been measured during the
#            experiment.  First dimension must be the same as the number
#            of full scans
#        s : str
#            stimulus block of interest
#        c : str
#            name of the control block to be used
#        offset : float
#            offset to apply at the beginning of an irritation phase
#        preset : float
#            offset to apply at the end of an irritation phase
#
#        Returns
#        -------
#        ndarray, shape_like(slice_time), dtype: float
#            The irritation vector indicating the kind of irritation under
#            which the subject was exposed while the particular slice has
#            been measured.
#        """
#        assert s != c, 'block names: {} and {c} must be different'.format(s,c)
#        assert s in self.names, 'name of stimulus block: {} does not exist'.format(s)
#        assert c in self.names, 'name of control block: {} does not exist'.format(c)
#
#        design = np.empty_like(slice_time)
#        design[...] = np.nan
#
#        onsets = self.onsets[c]
#        durations = self.durations[c]
#        srt = onsets + offset
#        end = onsets + durations - preset
#        mat = np.vstack((srt, end)).T
#
#        for x in iter(mat):
#            design[ (slice_time > x[0]) & (slice_time < x[1]) ] = 0
#
#        onsets = self.onsets[s]
#        durations = self.durations[s]
#        srt = onsets + offset
#        end = onsets + durations - preset
#        mat = np.vstack((srt, end)).T
#
#        for x in iter(mat):
#            design[ (slice_time > x[0]) & (slice_time < x[1]) ] = 1
#
#        return design

    def design(self, slice_timing, s, c, offset=0, preset=0):
        """
        Defines the irritation design vector

        Parameters
        ----------
        slice_timing : ndarray, shape (n,) or (n,m)
            times at which slices have been measured during the
            experiment.  First dimension must be the same as the number
            of full scans
        s : str
            stimulus block of interest
        c : str
            name of the control block to be used
        offset : float
            offset to apply at the beginning of an irritation phase
        preset : float
            offset to apply at the end of an irritation phase

        Returns
        -------
        ndarray, shape_like(slice_timing), dtype: float
            The irritation vector indicating the kind of irritation under
            which the subject was exposed while the particular slice has
            been measured.
        """
        assert s != c, 'block names: {} and {c} must be different'.format(s,c)
        assert s in self.names, 'name of stimulus block: {} does not exist'.format(s)
        assert c in self.names, 'name of control block: {} does not exist'.format(c)

        design = np.empty(slice_timing.shape + (2,) )
        design[...] = np.nan

        onsets = self.onsets[c]
        durations = self.durations[c]
        srt = onsets + offset
        end = onsets + durations - preset
        mat = np.vstack((srt, end)).T

        group = 0
        for x in iter(mat):
            block = (slice_timing > x[0]) & (slice_timing < x[1])
            design[block,0] = 0
            design[block,1] = group
            group += 1

        onsets = self.onsets[s]
        durations = self.durations[s]
        srt = onsets + offset
        end = onsets + durations - preset
        mat = np.vstack((srt, end)).T

        group = 0
        for x in iter(mat):
            block = (slice_timing > x[0]) & (slice_timing < x[1])
            design[block,0] = 1
            design[block,1] = group
            group += 1

        return design

    def number_of_onsets(self):
        return {name : len(self.onsets[name]) for name in self.names}

    def irritated_scans(self, slice_time, offset=0, preset=0):
        """
        Test whether blocks have been aquired during blocks of irritation

        Parameters
        ----------
        slice_time : ndarray, shape (n,) or (n,m), dtype: float

        Returns
        -------
        ndarray, shape_like(slice_time), dtype: bool
            If True, this scan has been aquired during a block of subject
            irritation.
        """
        onsets = self.onsets.items()

        srt = np.hstack([v + offset for k,v in onsets])
        end = np.hstack([v + irr.durations[k] - preset for k,v in onsets])
        mat = np.vstack((srt, end)).T

        irritated = np.zeros_like(slice_time).astype(bool)
        for x in iter(mat):
            irritated = irritated | (slice_time > x[0]) & (slice_time < x[1])

        return irritated

    def describe(self):
        """
        Give a description of the instance
        """
        description = """
        Type of irritation: block design
        Block number: {}
        Block names:  {}
        Onsets: {}
        Durations: {}
        Number of onsets per block: {}"""
        return description.format(
                self.number,
                self.names,
                self.durations,
                self.onsets,
                self.number_of_onsets(),
                )




# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Plot effect fields
------------------

Some handy functions for plotting fMRI images and effect fields.
"""

import numpy as np

import matplotlib
import matplotlib.pyplot as pt
import matplotlib.cm as cm

def shift_colour_map(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Shift a colour map

    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Parameters
    ----------
      cmap : cmap
          The matplotlib colormap to be altered
      start : float
          Offset from lowest point in the colormap's range.
          Defaults to 0.0 (no lower ofset). Should be between
          0.0 and `midpoint`.
      midpoint : float
          The new center of the colormap. Defaults to
          0.5 (no shift). Should be between 0.0 and 1.0. In
          general, this should be  1 - vmax/(vmax + abs(vmin))
          For example if your data range from -15.0 to +5.0 and
          you want the centre of the colormap at 0.0, `midpoint`
          should be set to  1 - 5/(5 + 15)) or 0.75
      stop : float
          Offset from highets point in the colormap's range.
          Defaults to 1.0 (no upper ofset). Should be between
          `midpoint` and 1.0.

    Returns
    -------
    cmap : The shifted colour map.
    '''
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False),
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    newcmap = matplotlib.colors.LinearSegmentedColormap(name, cdict)
    pt.register_cmap(cmap=newcmap)

    return newcmap

def cmap_shifted (dat, orig_cmap=cm.bwr):
    vmaxes = np.array((np.nanmin(dat), np.nanmax(dat)))
    vmax = max(vmaxes)
    vmin = min(vmaxes)
    midp = 1 - vmax/(vmax + abs(vmin))
    cmap_new = shift_colour_map(orig_cmap, midpoint=midp, name='shifted')
    return(cmap_new)

def picture(self, epi_code, nx, ny, slices=None, cmap=None,
        interpolation='none', background=None, mark_peak=False):

    if abs(epi_code) == 3:
        origin = 'lower'
        xlabel = 'Left -> Right'
        ylabel = 'Posterior -> Anterior'
        aspect = self.reference.aspect(1,0)
    if abs(epi_code) == 2:
        origin = 'lower'
        xlabel = 'Left -> Right'
        ylabel = 'Inferior -> Superior'
        aspect = self.reference.aspect(2,0)
    if abs(epi_code) == 1:
        origin = 'lower'
        xlabel = 'Posterior -> Anterior'
        ylabel = 'Inferior -> Superior'
        aspect = self.reference.aspect(2,1)

    imatrix, ixticks, iyticks, slices = self.flatten(epi_code,nx,ny,slices=slices)

    if (imatrix[np.isfinite(imatrix)] < 0).any():
        if cmap == None:
            cmap = cm.bwr
        cmap = cmap_shifted(imatrix, orig_cmap=cmap)
    else:
        if cmap == None:
            cmap = cm.magma

    if background:
        bmatrix, _, _, _ = background.flatten(epi_code,nx,ny,slices=slices)
        pt.imshow(bmatrix,interpolation=interpolation,cmap='Greys',origin=origin,aspect=aspect)

    im = pt.imshow(imatrix,interpolation=interpolation,cmap=cmap,origin=origin,aspect=aspect)
    #pt.colorbar()
    pt.box(on='off')
    pt.tick_params('off')
    pt.xlabel(xlabel)
    pt.ylabel(ylabel)

    ax = pt.axes()
    pt.setp(ax.get_yticklines(),visible=False)
    pt.setp(ax.get_xticklines(),visible=False)
    pt.setp(ax.get_xticklabels(),visible=False)
    pt.setp(ax.get_yticklabels(),visible=False)
    #ax.set_xticks(ixticks, minor=False)
    #ax.set_yticks(iyticks, minor=False)

    if mark_peak:
        peak  = np.nanargmax(imatrix)
        index = np.unravel_index(peak, imatrix.shape)
        pt.axvline(index[1], c='k', linewidth=.8)
        pt.axhline(index[0], c='k', linewidth=.8)

    return im, slices, imatrix

def contour(self, epi_code, nx, ny, slices=None, cmap=None,
        interpolation='none', background=None, mark_peak=False, **kwargs):

    if abs(epi_code) == 3:
        origin = 'lower'
        xlabel = 'Left -> Right'
        ylabel = 'Posterior -> Anterior'
        aspect = self.reference.aspect(1,0)
    if abs(epi_code) == 2:
        origin = 'lower'
        xlabel = 'Left -> Right'
        ylabel = 'Inferior -> Superior'
        aspect = self.reference.aspect(2,0)
    if abs(epi_code) == 1:
        origin = 'lower'
        xlabel = 'Posterior -> Anterior'
        ylabel = 'Inferior -> Superior'
        aspect = self.reference.aspect(2,1)

    imatrix, ixticks, iyticks, slices = self.flatten(epi_code,nx,ny,slices=slices)

    if (imatrix[np.isfinite(imatrix)] < 0).any():
        if cmap == None:
            cmap = cm.bwr
        cmap = cmap_shifted(imatrix, orig_cmap=cmap)
    else:
        if cmap == None:
            cmap = cm.magma

    if background:
        bmatrix, _, _, _ = background.flatten(epi_code,nx,ny,slices=slices)
        pt.imshow(bmatrix,interpolation=interpolation,cmap='Greys',origin=origin,aspect=aspect)

    cs=pt.contour(imatrix, **kwargs)
    pt.box(on='off')
    pt.tick_params('off')
    pt.xlabel(xlabel)
    pt.ylabel(ylabel)

    ax = pt.axes()
    pt.setp(ax.get_yticklines(),visible=False)
    pt.setp(ax.get_xticklines(),visible=False)
    pt.setp(ax.get_xticklabels(),visible=False)
    pt.setp(ax.get_yticklabels(),visible=False)
    #ax.set_xticks(ixticks, minor=False)
    #ax.set_yticks(iyticks, minor=False)
    #pt.colorbar()

    if mark_peak:
        peak  = np.nanargmax(imatrix)
        index = np.unravel_index(peak, imatrix.shape)
        pt.axvline(index[1], c='k', linewidth=.8)
        pt.axhline(index[0], c='k', linewidth=.8)

    return cs, slices, imatrix



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

A population map is a diffeomorphism from a population space to the
subject reference space of a FMRI session.

"""

from .name import Identifier

from .affines import Affine, isclose, from_cartesian

from .diffeomorphisms import Diffeomorphism, Image, Identity, AffineTransformation

import pickle

import numpy as np

#######################################################################
#
# PopulationMap
#
#######################################################################

class PopulationMap:
    """
    A population map is a diffeomorphism from a population space to the
    reference space of a FMRI session.

    Parameters
    ----------
    diffeomorphism : subclass of Diffeomorphism
        The diffeomorphism ψ that maps from each point in the data field
        of the template brain to the subject reference space of an FMRI
        session.
    template : None, Image or ndarray
        A template image in population space.
    template_mask : None, Image or ndarray
        A template mask image in population space. By default the
        template mask image will be used in many operations. The idea:
        You provide a detailed image, say also containing skull, in
        template but only the brain as in the template mask image.
    target : None, Image or ndarray
        A target image in population space.

    Notes
    -----
    If template is an Image, it will be checked whether its reference
    agrees with the reference of the diffeomorphism.  If template is an
    ndarray, it must be equal to the shape stored with the
    diffeomorphism, as it will then be assumed that they share the same
    reference: the template's reference will be set identical to the
    reference of the diffeomorphism.

    If the software which has been used to fit the diffeomorphism
    creates a warped image of its input, it may be stored in `target`.
    """

    def __init__(self, diffeomorphism, template=None,
            template_mask=None, target=None, metadata=None):

        assert issubclass(type(diffeomorphism), Diffeomorphism), \
                'diffeomorphism must be a subclass of Diffeomorphism'
        assert type(diffeomorphism.nb) is Identifier, \
                'nb of diffeomorphism must be Identifier'

        self.diffeomorphism = diffeomorphism
        self.name = self.diffeomorphism.nb

        if template:
            self.set_template(template=template)

        if template_mask:
            self.set_template_mask(template_mask=template_mask)

        if target:
            self.set_target(target=target)

        self.metadata = metadata

    def set_template(self, template):
        """
        Set or reset template image

        Parameters
        ----------
        template : Image or ndarray
            The data array.

        Notes
        -----
        If template is an Image, it will be checked whether its
        reference agrees with the reference of the diffeomorphism.  If
        template is an ndarray, it must be equal to the shape stored
        with the diffeomorphism, as it will then be assumed that they
        share the same reference: the template's reference will be set
        identical to the reference of the diffeomorphism.
        """
        assert template.shape == self.diffeomorphism.shape, \
                'shapes of template and diffeomorphism must match'

        if type(template) is Image:
            assert isclose(template.reference, self.diffeomorphism.reference), \
                    'references of template and diffeomorphism must match'
            template.reference = self.diffeomorphism.reference
            self.template = template
        else:
            self.template = Image(
                    reference=self.diffeomorphism.reference,
                    data=template,
                    name=self.diffeomorphism.vb)

    def set_template_mask(self, template_mask):
        """
        Set or reset template mask image

        Parameters
        ----------
        template_mask : Image or ndarray
            The data array.

        Notes
        -----
        If template_mask is an Image, it will be checked whether its
        reference agrees with the reference of the diffeomorphism.  If
        template_mask is an ndarray, it must be equal to the shape
        stored with the diffeomorphism, as it will then be assumed that
        they share the same reference: the template_mask's reference
        will be set identical to the reference of the diffeomorphism.
        """
        assert template_mask.shape == self.diffeomorphism.shape, \
                'shapes of template_mask and diffeomorphism must match'

        if type(template_mask) is Image:
            assert isclose(template_mask.reference, self.diffeomorphism.reference), \
                    'references of template_mask and diffeomorphism must match'
            template_mask.reference = self.diffeomorphism.reference
            self.template_mask = template_mask
        else:
            self.template_mask = Image(
                    reference=self.diffeomorphism.reference,
                    data=template_mask,
                    name=self.diffeomorphism.vb)

    def set_target(self, target):
        """
        Set or reset target image

        Parameters
        ----------
        target : Image or ndarray, shape (x,y,z)
            The data array.

        Notes
        -----
        If target is an Image, it will be checked whether its
        reference agrees with the reference of the diffeomorphism.  If
        target is an ndarray, it must be equal to the shape stored
        with the diffeomorphism, as it will then be assumed that they
        share the same reference: the target's reference will be set
        identical to the reference of the diffeomorphism.
        """
        assert target.shape == self.diffeomorphism.shape, \
                'shapes of target and diffeomorphism must match'

        if type(target) is Image:
            assert isclose(target.reference, self.diffeomorphism.reference), \
                    'references of target and diffeomorphism must match'
            target.reference = self.diffeomorphism.reference
            self.target = target
        else:
            self.target = Image(
                    reference=self.diffeomorphism.reference,
                    data=target,
                    name=self.diffeomorphism.vb)

    def describe(self):
        description = """
            Available images in population space
            ------------------------------------
            Template:      {}
            Template mask: {}
            Target:        {}
        """
        return description.format(
                hasattr(self, 'template'),
                hasattr(self, 'template_mask'),
                hasattr(self, 'target'),
                )

    def save(self, file, **kwargs):
        """
        Save instance to disk

        Parameters
        ----------
        file : str
            A file name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)

#######################################################################
#
# Functions for creating canonical population maps
#
#######################################################################

def pmap_scanner(session):
    """
    The identity map

    Sets the population space equal to the coordinate system of the
    scanner and the population map equal to the identity, and the
    template grid to native resolution.

    Parameters
    ----------
    session : Session
        A FMRI session

    Notes
    -----
    When only studying a single subject, it makes perfect sense to
    set the population space equal to an isometric image of the
    subject's head, i.e. to a population space which preserves distances
    with respect to the subject.

    In other words, for a single subject, the population space can be
    set equal to the subject reference space :math:`R` or to
    :math:`ρ_s[R]` for some scan reference :math:`ρ_s`. This function
    will do the former and sets the resolution of the template equal to
    the resolution of the acquisition grid. (When fitting on this grid,
    the signal model is estimated in *native* resolution).

    It only makes sense, though, to set the population space equal to
    the coordinate system of the FMRI session, if this space is indeed
    *close* to the subject reference space of the FMRI session This may
    be archived by calling the :func:`reset_reference_space` attribute
    of the session first. .

    See also
    --------
    `pmap_reference`
    `pmap_scan`
    """
    diffeomorphism = Identity(
            reference=session.reference,
            shape=session.shape,
            vb='scanner',
            nb=session.name)

    return PopulationMap(diffeomorphism)

def pmap_reference(session, resolution=2.):
    """
    The identity map

    Sets the population space equal to the coordinate system of the
    scanner and the population map equal to the identity, and the
    template grid to given resolution.

    Parameters
    ----------
    session : Session
        A FMRI session
    resolution : float
        Resolution in milli meter, default 2.

    Notes
    -----
    When only studying a single subject, it makes perfect sense to
    set the population space equal to an isometric image of the
    subject's head, i.e. to a population space which preserves distances
    with respect to the subject.

    In other words, for a single subject, the population space can be
    set equal to the subject reference space :math:`R` or to
    :math:`ρ_s[R]` for some scan reference :math:`ρ_s`. This function
    will do the former and sets the resolution to the given value.

    It only makes sense, though, to set the population space equal to
    the coordinate system of the FMRI session, if this space is indeed
    *close* to the subject reference space of the FMRI session This may
    be archived by calling the :func:`reset_reference_space` attribute
    of the session first. .

    See also
    --------
    :func:`pmap_scanner`
    :func:`pmap_scan`
    """

    ref = session.reference
    mat = ref.affine[:3,:3] / (ref.resolution() / resolution)

    reference = from_cartesian(mat, ref.affine[:3,3])
    shape = tuple(np.ceil(reference.inv().apply(ref.apply_to_index(session.shape))).astype(int))

    diffeomorphism = Identity(
            reference=reference,
            shape=shape,
            vb='reference',
            nb=session.name)

    return PopulationMap(diffeomorphism)

def pmap_scan(session, scan_cycle):
    """
    Pick a scan reference as the population map

    Parameters
    ----------
    session : Session
        Session instance
    scan_cycle : int
        Reference scan cycle.

    Notes
    -----
    When only studying a single subject, it makes perfect sense to
    set the population space equal to an isometric image of the
    subject's head, i.e. to a population space which preserves distances
    with respect to the subject.

    In other words, for a single subject, the population space can be
    set equal to the subject reference space :math:`R` or to
    :math:`ρ_s[R]` for some scan reference :math:`ρ_s`. This function
    will do the latter for the given scan, and sets the resolution of
    the template equal to the resolution of the acquisition grid. (When
    fitting on this grid, the signal model is estimated in *native*
    resolution).

    See also
    --------
    :func:`pmap_reference`
    :func:`pmap_scanner`
    """
    diffeomorphism = AffineTransformation(
            reference=session.reference,
            affine=session.scan_inverses[scan_cycle],
            shape=session.shape,
            vb='reference',
            nb=session.name)

    return PopulationMap(diffeomorphism)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Prune

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('--fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='input file;' + hp.sfit)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('--pruned',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='input file;' + hp.sfit)

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fsl4prune.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('--protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--datetime',
            help=hp.datetime)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

    parser.add_argument('--population-space',
            default='reference',
            help=hp.population_space)

    parser.add_argument('--scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help=hp.scale_type)

########################################################################
# Arguments specific for the application of masks
########################################################################

    handling_df_cutoff = parser.add_mutually_exclusive_group()

    handling_df_cutoff.add_argument('-p', '--proportion',
            type=float,
            help="""estimates which degrees of freedom are below the
            proportional threshold of the degrees of freedom in the
            effect field estimate are set to null.""")

    handling_df_cutoff.add_argument('-t', '--threshold',
            type=int,
            help="""estimates which degrees of freedom are below the
            threshold of the degrees of freedom in the effect field
            estimate are set to null.""")

########################################################################
# Miscellaneous
########################################################################

    parser.add_argument('-f', '--force',
            action='store_true',
            help="""Overwrite mask if it already exits""")

    parser.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

########################################################################
# Multiprocessing
########################################################################

    parser.add_argument('-j', '--cores',
            type=int,
            help=hp.cores)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...load import load_result

from ...name import Identifier

from ...protocol import layout_dummy, layout_sdummy

from ...smodel import Result

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

import nibabel as ni

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args, fall_back=args.fit)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_sdummy(df_layout, 'filename',
            template=args.fit,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    ####################################################################
    # Apply wrapper
    ####################################################################

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        wrapper(name                  = name,
                df                    = df,
                index                 = r.Index,
                filename              = r.filename,
                verbose               = args.verbose,
                force                 = args.force,
                vb                    = args.population_space,
                threshold_df          = args.threshold,
                proportion_df         = args.proportion,
                )

    it =  df_layout.itertuples()

    if len(df_layout) > 1 and ((args.cores is None) or (args.cores > 1)):
        try:
            pool = ThreadPool(args.cores)
            results = pool.map(wm, it)
            pool.close()
            pool.join()
        except Exception as e:
            pool.close()
            pool.terminate()
            print('Pool execution has been terminated')
            print(e)
        finally:
            pass
    else:
        try:
            print('Process protocol entries sequentially')
            for r in it:
                wm(r)
        finally:
            pass

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

########################################################################

def wrapper(name, df, index, filename, verbose, force, vb, threshold_df, proportion_df):

    result = load_result(filename, name, df, index, vb, verbose)
    if df.ix[index,'valid'] == False:
        return

    if verbose > 1:
        print('{}: Description of the fit:'.format(result.name.name()))
        print(result.describe())

    if hasattr(result.population_map, 'template_mask') and \
        (result.population_map.template_mask is not None) and not force:
        if verbose:
            print('{}: Template mask already exits. Use -f/--force to overwrite'.format(name.name()))
        return

    gf = result.get_field('df', 'other')

    if proportion_df:
        threshold_df = int(proportion_df * np.nanmax(gf.data))

    if verbose:
        print('{}: Lower df threshold: {:d}'.format(name.name(), threshold_df))

    result.population_map.set_template_mask(gf.data >= threshold_df)

    if verbose:
        print('{}: Save: {}'.format(name.name(), filename))

    try:
        result.save(filename)
        df.ix[index,'locked'] = False
    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to write: {}'.format(name.name(), filename))
        print('{}: Exception: {}'.format(name.name(), e))



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

from .lock import Lock

from .name import Identifier

from .affines import isclose

from .irritation import Block

from .session import Session

from .reference import ReferenceMaps

from .pmap import PopulationMap

from .smodel import SignalModel, Result

from pandas import DataFrame

import numpy as np

import pickle

def load(file):
    """
    Load instances from disk

    This will load a call instances from disk.

    Parameters
    ----------
    file : str
        File name.
    """
    with open(file, 'rb') as input:
        self = pickle.load(input)

    return self

def load_block_irritation(file, name, df, index, verbose=True):
    """
    Load a irritation file

    This will performe some tests whether the correct file has been
    loaded from disk.  This function is frequently used in the command
    line interface for fmristats.

    Parameters
    ----------
    file : str
    name : Identifier
    df : DataFrame
    verbose : bool

    Returns
    -------
    Irritation
    """
    assert type(name) is Identifier, 'name must be Identifier'
    assert type(df) is DataFrame, 'df must be DataFrame'

    try:
        irritation = load(file)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file))
    except Exception as e:
        print('{}: Unable to read: {}'.format(name.name(), file))
        #print('{}: Exception: {}'.format(name.name(), e))
        df.ix[index,'valid'] = False
        return

    if type(irritation) is Block and name.is_equal(irritation.name):
        return irritation
    elif type(irritation) is Lock:
        df.ix[index,'valid'] = False
        print('{}: Irritation is currently locked'.format(name.name(), file))
        return irritation
    else:
        df.ix[index,'valid'] = False
        print("""{}: Irritation does not match:
        Expected: {}
        Found:    {}""".format(name.name(), name.name(True), irritation.name.name(True)))
        return irritation

def load_session(file, name, df, index, verbose=True):
    """
    Load a session file

    This will performe some tests whether the correct file has been
    loaded from disk.  This function is frequently used in the command
    line interface for fmristats.

    Parameters
    ----------
    file : str
    name : Identifier
    df : DataFrame
    verbose : bool

    Returns
    -------
    Session
    """
    assert type(name) is Identifier, 'name must be Identifier'
    assert type(df) is DataFrame, 'df must be DataFrame'

    try:
        session = load(file)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file))
    except Exception as e:
        print('{}: Unable to read: {}'.format(name.name(), file))
        #print('{}: Exception: {}'.format(name.name(), e))
        df.ix[index,'valid'] = False
        return

    if type(session) is Session and name.is_equal(session.name):
        return session
    elif type(session) is Lock:
        df.ix[index,'valid'] = False
        print('{}: Session is currently locked'.format(name.name(), file))
        return session
    else:
        df.ix[index,'valid'] = False
        print("""{}: Session does not match:
        Expected: {}
        Found:    {}""".format(name.name(), name.name(True), session.name.name(True)))
        return session

def load_refmaps(file, name, df, index, verbose=True):
    """
    Load a file

    This will performe some tests whether the correct file has been
    loaded from disk.  This function is frequently used in the command
    line interface for fmristats.

    Parameters
    ----------
    file : str
    name : Identifier
    df : DataFrame
    index : int
    verbose : bool

    Returns
    -------
    ReferenceMaps
    """
    assert type(name) is Identifier, 'name must be Identifier'
    assert type(df) is DataFrame, 'df must be DataFrame'

    try:
        reference_maps = load(file)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file))
    except Exception as e:
        print('{}: Unable to read: {}'.format(name.name(), file))
        #print('{}: Exception: {}'.format(name.name(), e))
        df.ix[index,'valid'] = False
        return

    if type(reference_maps) is ReferenceMaps and name.is_equal(reference_maps.name):
        return reference_maps
    elif type(reference_maps) is Lock:
        df.ix[index,'valid'] = False
        print('{}: ReferenceMaps is currently locked'.format(name.name(), file))
        return reference_maps
    else:
        df.ix[index,'valid'] = False
        print("""{}: ReferenceMaps does not match:
        Expected: {}
        Found:    {}""".format(name.name(), name.name(True), reference_maps.name.name(True)))
        return reference_maps

def load_result(file, name, df, index, vb, verbose=True):
    """
    Load a file

    This will performe some tests whether the correct file has been
    loaded from disk.  This function is frequently used in the command
    line interface for fmristats.

    Parameters
    ----------
    file : str
    name : Identifier
    df : DataFrame
    index : int
    verbose : bool

    Returns
    -------
    Result
    """
    assert type(name) is Identifier, 'name must be Identifier'
    assert type(df) is DataFrame, 'df must be DataFrame'

    try:
        result = load(file)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file))
    except Exception as e:
        print('{}: Unable to read: {}'.format(name.name(), file))
        #print('{}: Exception: {}'.format(name.name(), e))
        df.ix[index,'valid'] = False
        return

    if type(result) is Result and name.is_equal(result.population_map.name):
        if vb is None:
            if np.isnan(result.statistics).all():
                df.ix[index,'valid'] = False
                print('{}: Result is completely empty'.format(name.name()))
                return result
            else:
                return result
        else:
            if vb == result.population_map.diffeomorphism.vb:
                if np.isnan(result.statistics).all():
                    df.ix[index,'valid'] = False
                    print('{}: Result is completely empty'.format(name.name()))
                    return result
                else:
                    return result
            else:
                df.ix[index,'valid'] = False
                print("""{}: Result does not match:
                Expected: {}
                Found:    {}""".format(name.name(), vb, population_map.diffeomorphism.vb))
                return result
    elif type(result) is Lock:
        df.ix[index,'valid'] = False
        print('{}: Result is currently locked'.format(name.name()))
        return result
    else:
        df.ix[index,'valid'] = False
        print("""{}: Result does not match:
        Expected: {}
        Found:    {}""".format(name.name(), name.name(True), result.name.name(True)))
        return result

def load_population_map(file, name, df, index, vb, verbose=True):
    """
    Load a file

    This will perform some tests whether the correct file has been
    loaded from disk.  This function is frequently used in the command
    line interface for fmristats.

    Parameters
    ----------
    file : str
    name : Identifier
    df : DataFrame
    index : int
    verbose : bool

    Returns
    -------
    PopulationMap
    """
    assert type(name) is Identifier, 'name must be Identifier'
    assert type(df) is DataFrame, 'df must be DataFrame'

    try:
        population_map = load(file)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file))
    except Exception as e:
        print('{}: Unable to read: {}'.format(name.name(), file))
        #print('{}: Exception: {}'.format(name.name(), e))
        df.ix[index,'valid'] = False
        return

    if type(population_map) is PopulationMap and name.is_equal(population_map.diffeomorphism.nb):
        if vb is None:
            return population_map
        else:
            if vb == population_map.diffeomorphism.vb:
                return population_map
            else:
                df.ix[index,'valid'] = False
                print("""{}: PopulationMap does not match:
                Expected: {}
                Found:    {}""".format(name.name(), vb, population_map.diffeomorphism.vb))
                return population_map
    elif type(population_map) is Lock:
        df.ix[index,'valid'] = False
        print('{}: PopulationMap is currently locked'.format(name.name(), file))
        return population_map
    else:
        df.ix[index,'valid'] = False
        print("""{}: PopulationMap does not match:
        Expected: {}
        Found:    {}""".format(name.name(), name.name(True), population_map.name.name(True)))
        return population_map



"""
Euler angels
------------

Code for calculating Euler angels from rotation matrices
"""

# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# The following code has been almost identically taken from:
#   http://www.learnopencv.com/rotation-matrix-to-euler-angles/

import numpy as np
import math

def is_rotation_matrix(a):
    """
    Checks if a matrix is a valid rotation matrix.
    """
    shouldBeIdentity = np.dot(np.transpose(a), a)
    I = np.identity(3, dtype = a.dtype)
    n = np.linalg.norm(I - shouldBeIdentity)
    return n < 1e-6

def rotation_matrix_to_euler_angles(a) :
    """
    Calculates rotation matrix to Euler angles

    The result is the same as MATLAB except the order
    of the Euler angles (x and z are swapped).
    """
    assert(is_rotation_matrix(a))
    sy = math.sqrt(a[0,0] * a[0,0] +  a[1,0] * a[1,0])
    singular = sy < 1e-6

    if  not singular :
        x = math.atan2(a[2,1] , a[2,2])
        y = math.atan2(-a[2,0], sy)
        z = math.atan2(a[1,0], a[0,0])
    else :
        x = math.atan2(-a[1,2], a[1,1])
        y = math.atan2(-a[2,0], sy)
        z = 0

    return np.array([x, y, z])



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

from .name import Identifier

from .affines import Affine, Affines

from .diffeomorphisms import Image, Diffeomorphism

from .irritation import Irritation, Block

from .session import Session

from .reference import ReferenceMaps

from .pmap import PopulationMap

from .smodel import SignalModel, Result

from .sample import Sample

from .pmodel import MetaResult, PopulationModel

from .tau import tau

from .load import load



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Methods for outlier detection.

"""

import numpy as np

from scipy.stats.distributions import t

def grubbs_test(xvec, sgnf, null='equal'):
    """
    The Grubbs' test for outlyer detection

    Parameters
    ----------
    xvec : ndarray, shape (n,)
    sgnf : float
    null : str

    Notes
    -----
    Specification of the null hypothesis can either be 'equal'
    (default), 'greater', or 'less'. Only 'equal' is implemented.

    Missigs (nan values) are ignored in the test.
    """
    assert np.not_equal(np.nanvar(xvec), 0), 'variance of observations (xvec) is equal to 0'

    n = sum(~np.isnan(xvec))
    m = np.nanmean(xvec)
    s = np.sqrt(np.nanvar(xvec))
    candidates = abs(xvec - m)
    gstat = np.nanmax(candidates) / s
    tstat = t.ppf(q=1-(sgnf / (2*n)), df=n-2)**2
    crval = ((n-1)/np.sqrt(n)) * np.sqrt(tstat / (n-2+tstat))
    if (gstat > crval):
        return np.nanargmax(candidates)
    else:
        return None

def grubbs(xvec, sgnf, inplace=False):
    """
    Iterate Grubbs' test

    Parameters
    ----------
    xvec : ndarray, shape (n,)
    sgnf : float
    inplace : bool

    Notes
    -----
    Missigs (nan values) are set to True.
    """
    if not inplace:
        xvec = xvec.copy()
    args = np.repeat(False, len(xvec))
    args [ np.isnan(xvec) ] = True
    while (sum(~np.isnan(xvec)) > 6 & np.not_equal(np.nanvar(xvec), 0)):
        arg = grubbs_test(xvec, sgnf)
        if arg == None:
            break
        xvec [arg] = np.nan
        args [arg] = True
    return xvec, args



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Query covariate files

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('protocol',
            help=hp.protocol)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('-o', '--output',
            help="""output file""")

########################################################################
# Arguments specific for using the protocol API
########################################################################

    #to_process = parser.add_argument_group(
    #        """specifying the covariate entries to process""",
    #        """Arguments which give control which entries to
    #        process.""")

    #to_process.add_argument('--cohort',
    #        help=hp.cohort)

    #to_process.add_argument('--id',
    #        type=int,
    #        nargs='+',
    #        help=hp.j)

########################################################################
# Arguments specific for using fmriprotocol
########################################################################

    parser.add_argument('--covariate',
            help="""update field 'valid' in the protocol file such that
            only entries which are valid in the protocol *and* in
            COVARIATE file are marked as valid entries.""")

    parser.add_argument('--update',
            help="""update field 'valid' in the protocol file with the
            entries in the file UPDATE.""")

    parser.add_argument('--merge',
            help="""merge with covariate file MERGE.""")

    parser.add_argument('--query',
            help="""apply a QUERY to the data frame.""")

    parser.add_argument('--head',
            type=int,
            help="""print the first HEAD number of lines""")

    parser.add_argument('--to-csv',
            help="""export protocol file to CSV""")

########################################################################
# Miscellaneous
########################################################################

    misc = parser.add_argument_group(
            """miscellaneous""")

    misc.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('protocol'))

    misc.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import filter_df

from ...name import Identifier

import pandas as pd

import sys

import os

from os.path import isfile, isdir, join

########################################################################

def call(args):
    try:
        df = pd.read_pickle(args.protocol)
        if args.verbose:
            print('Read: {}'.format(args.protocol))
    except Exception as e:
        print('Unable to read protocol file {}'.format(args.protocol))
        print('Exception: {}'.format(e))
        sys.exit()

    if args.verbose:
        valid = df.groupby(['cohort','valid']).id.agg(['count'])
        print('Number of entries in the list: {}'.format(len(df)))
        print("""\nOverview""")
        print(valid)
        print()

    if args.query:
        df = df.query(args.query).copy()

        if args.verbose:
            valid = df.groupby(['cohort','valid']).id.agg(['count'])
            print('Number of entries in the list after --query: {}'.format(len(df)))
            print("""\nOverview""")
            print(valid)
            print()

    if args.head:
        if args.verbose:
            print()
            print("""-------------------------------------------------------------""".format(args.head))
        print(df.head(args.head))
        print()

    if args.output:
        if not isfile(args.output) or args.force:
            if args.verbose:
                print('Save: {}'.format(args.output))
            df.to_pickle(args.output)
        else:
            print('File already exists: {}, use --force to overwrite'.format(args.output))

    if args.to_csv:
        if not isfile(args.to_csv) or args.force:
            if args.verbose:
                print('Save: {}'.format(args.to_csv))
            df.to_csv(args.to_csv, index=True)
        else:
            print('File already exists: {}, use --force to overwrite'.format(args.to_csv))



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Extract information from instance files and print information to
standard output

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

    parser.add_argument('files', nargs='+', help='input files')

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ...lock import Lock

from ...affines import Affine

from ...diffeomorphisms import Image, Diffeomorphism

from ...irritation import Irritation, Block

from ...session import Session

from ...reference import ReferenceMaps

from ...pmap import PopulationMap

from ...smodel import SignalModel, Result

from ... import load

from ...sample import Sample

#from ...pmodel import PopulationModel, MetaResult

import pandas as pd

from pandas import DataFrame

import numpy as np

########################################################################

def call(args):
    for f in args.files:
        print_info(load(f), f)

def print_info(x, f):
    if type(x) is Image:
        print('{}: image file'.format(f))
        print(x.describe())

    if type(x) is Session:
        print('{}: session file'.format(f))
        print(x.name.describe())
        print(x.describe())
        print(x.irritation.describe())
        print(x.reference.describe())

    if type(x) is Irritation:
        print('{}: irritation file'.format(f))
        print(x.describe())

    if type(x) is Block:
        print('{}: block irritation file'.format(f))
        print(x.describe())

    if type(x) is ReferenceMaps:
        print('{}: reference maps'.format(f))
        print(x.name.describe())
        print(x.describe())

    if type(x) is PopulationMap:
        print('{}: population map'.format(f))
        print(x.diffeomorphism.describe())
        print(x.describe())

    if type(x) is Result:
        print('{}: fit of a signal model'.format(f))
        print(x.describe())
        print(x.population_map.diffeomorphism.describe())
        print(x.population_map.describe())

    if type(x) is Lock:
        print('{}: lock'.format(f))
        print(x.describe())

    if type(x) is Sample:
        print('{}: sample file'.format(f))
        print(x.describe())

    if type(x) is DataFrame:
        if 'cohort' in x.columns:
            x.cohort.cat.remove_unused_categories(inplace=True)
            if 'valid' in x.columns:
                if 'paradigm' in x.columns:
                    x.paradigm.cat.remove_unused_categories(inplace=True)
                    valid = x.groupby(['cohort','paradigm','valid']).id.agg(['count'])
                else:
                    valid = x.groupby(['cohort','valid']).id.agg(['count'])
        else:
            valid = None

        print("""{}: protocol file

No. of entries: {:d}

First five entries:
{}

Data types:
{}

Valid entries:
{}""".format(f, len(x), x.head(), x.dtypes, valid))




# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Defines a class for FMRI sessions, which allows to define the
experimental setting and design, and to set up necessary parameters
needed for the statistical analysis of single subject data.

"""

from .name import Identifier

from .affines import Affine, isinverse

from .irritation import Irritation

from .observations import create_slice_time_matrix

from .filters import fit_foreground

from .diffeomorphisms import Image

import numpy as np

import pickle

#######################################################################
#######################################################################
#
# Describe the experimental design of an FMRI session
#
#######################################################################
#######################################################################

def fmrisetup(session, irritation):
    """
    If a session instance was created using the basic information in a
    Nifti1 file, this will complete the set-up of the session by
    providing the rest of the necessary FMRI session information:
    irritation design (a.k.a paradigm).

    Parameters
    ----------
    reset_reference_space : bool, default True
    """
    session.set_slice_time()
    session.set_irritation(irritation)

#######################################################################

class Session:
    """
    A FMRI session

    Collects attributes which reasonably define a FMRI
    session, e.g., experimental design, the on- and offsets of the
    blocks or events of the paradigm, the observed FMRI signals.

    Parameters
    ----------
    name : Identifier
        The name or of the FMRI experiment
    data : array, shape (M,N,O,P)
        The data of observations, where M is the number of scan
        cycles
    epi_code : int
        Coded EPI sequence, either one of -3, -2, -1, 1, 2, 3
    spacial_resolution : ndarray, shape (3,), dtype: float
        Spacial resolution of the image
    temporal_resolution : float
        Temporal resolution of the image
    reference : Affine or ndarray, shape (4,4), dtype: float
        The scanner reference
    """

    def __init__(self, name, data, epi_code, spacial_resolution,
            temporal_resolution, reference):
        assert type(name) is Identifier, 'name must be of type Identifier'
        assert epi_code <=  3, 'epi_code must be ≤ 3'
        assert epi_code >= -3, 'epi_code must be ≥-3'
        assert epi_code !=  0, 'epi_code must be different from 0'

        self.name = name
        self.raw  = data      # will always contain the raw data
        self.data = self.raw  # will (one day) become a numpy masked array

        self.epi_code = epi_code
        self.ep = abs(epi_code)-1

        self.spacial_resolution = spacial_resolution
        self.temporal_resolution = temporal_resolution

        self.shape = self.raw.shape[1:]
        self.numob = self.raw.shape[0]

        if type(reference) is Affine:
            self.reference = reference
        else:
            self.reference = Affine(reference)

    ####################################################################
    # Functions that act on the data or extract information from it
    ####################################################################

    def set_slice_time(self, slice_time=None, interleaved=False):
        """
        Set up the slice time information

        Parameters
        ----------
        slice_time : ndarray of float, shape (n,) or (n,x,y,z)
            The time vector indicating the point in time at which a
            given slice of the acquisition grid has been measured.  If
            slice_time has shape (n,), all slices of a scan cycle are
            assumed to have been measured during the same time.
        """
        if slice_time is None:
            slice_time = create_slice_time_matrix(
                shape               = self.data.shape,
                epi_code            = self.epi_code,
                temporal_resolution = self.temporal_resolution,
                interleaved         = False)

        shape_test = (self.numob, self.shape[self.ep])
        assert slice_time.shape == shape_test, \
                'slice times have shape: {}, expected {}'.format(
                        slice_time.shape, shape_test)

        self.slice_time = slice_time

    def set_irritation(self, irritation):
        """
        Set up the irritation
        """
        assert issubclass(type(irritation), Irritation) or type(irritation) is Irritation, \
                'type must be Irritation or a subclass if the same'
        self.irritation = irritation

    ####################################################################
    # Foreground detection
    ####################################################################

    def fit_foreground(self):
        self.data = self.raw.astype(float).copy()
        self.thresholds = fit_foreground(self.data, ep=self.ep)

    def set_foreground(self, foreground, is_mask=False):
        assert foreground.shape == data.shape, 'foreground and data not of equal shape'
        assert type(is_mask) is bool, 'is_mask must be bool'

        if is_mask:
            self.data = self.raw.astype(float).copy()
            self.data [ ~(foreground > 0) ] = np.nan
        else:
            self.data = foreground

    ###################################################################
    # Descriptive statistics of this session
    ###################################################################

    def descriptive_statistics(self):
        """
        Descriptive statistics for assessing the quality of the
        MR-signal on scan cycle level.
        """
        #TODO: give nicer description of what this function does
        ct, pt = self.thresholds
        dat = np.moveaxis(self.raw, (0, self.ep+1), (-1,-2))

        foreground = (dat >= pt) | (dat >= ct)
        background = (dat <  pt) | (dat >= ct)

        # Mean signal in the foreground
        muf = dat [ foreground ].mean()

        # Mean signal in the background
        mub = dat [ background ].mean()

        # Mean background to foreground signal
        mfb = mub / muf

        # Standard deviation in the foreground
        sdf = dat [ foreground ].std()

        # Standard deviation in the background
        sdb = dat [ background ].std()

        # foreground noise to mean signal foreground (foreground coefficient of variation)
        cvf = sdf / muf

        # background noise to mean signal foreground (background coefficient of variation)
        cvb = sdb / muf

        return muf, mub, mfb, sdf, sdb, cvf, cvb

    def descriptive_statistics_detailed(self):
        """
        Descriptive statistics for assessing the quality of the
        MR-signal on scan level.
        """
        #TODO: give nicer description of what this function does
        ct, pt = self.thresholds
        dat = np.moveaxis(self.raw, (0, self.ep+1), (-1,-2))

        background = (dat <  pt)
        foreground = ~background

        sumstats = np.zeros((pt.shape+(7,)))
        sumstats.shape

        fore = dat.astype(float)
        back = fore.copy()
        fore [ background ] = np.nan
        back [ foreground ] = np.nan

        muf = np.nanmean(fore, axis=(0,1)) # mean foreground signal
        mub = np.nanmean(back, axis=(0,1)) # mean background signal
        mfb = mub / muf                    # mean background to mean foreground signal
        sdf = np.nanstd(fore, axis=(0,1))  # standard deviation of foreground signal
        sdb = np.nanstd(back, axis=(0,1))  # standard deviation of foreground signal
        cvf = sdf / muf                    # foreground coefficient of variation
        cvb = sdb / muf                    # background coefficient of variation

        return muf, mub, mfb, sdf, sdb, cvf, cvb

    def describe(self):
        description = """
        EPI code: {:d}"""
        return description.format(
                self.epi_code,
                self.spacial_resolution,
                self.temporal_resolution,
                )

    #######################################################################
    # Save instance to disk
    #######################################################################

    def save(self, file, **kwargs):
        """
        Save model instance to disk

        This will save the current model instance to disk for later use.

        Parameters
        ----------
        file : str
            File name.
        """
        with open(file, 'wb') as output:
            pickle.dump(self, output, **kwargs)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

import numpy as np

from datetime import datetime

class Identifier:
    """
    Identifier

    Parameters
    ----------
    cohort : str
    j : int
    datetime : datetime
    paradigm : str
    """

    template = '{}-{:04d}-{}'
    template_long = '{}-{:04d}-{}-{}'
    template_strftime = '%Y-%m-%d-%H%M'

    def __init__(self, cohort, j, datetime, paradigm):
        assert type(cohort) == str, 'cohort must be str'
        assert (type(j) == int) or (type(j) == np.int64), 'j must be integer'
        assert type(paradigm) == str, 'paradigm must be str'
        self.cohort = cohort
        self.j = j
        self.datetime = datetime
        self.paradigm = paradigm

    def is_equal(self, x):
        assert type(x) is Identifier

        if self.cohort   != x.cohort:
            return False
        if self.j        != x.j:
            return False
        if self.datetime != x.datetime:
            return False
        if self.paradigm != x.paradigm:
            return False

        return True

    def name(self, long=False):
        if long:
            return self.template_long.format(
                    self.cohort,
                    self.j,
                    datetime.strftime(self.datetime, self.template_strftime),
                    self.paradigm)
        else:
            return self.template.format(self.cohort, self.j, self.paradigm)

    def describe(self):
        description = """
        Cohort:   {}
        Subject:  {}
        Date:     {}
        Paradigm: {}"""
        return description.format(
                self.cohort,
                self.j,
                datetime.strftime(self.datetime, self.template_strftime),
                self.paradigm,
                )



# Copyright 2016-2018 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Command line tool to create a fmristats session instance

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('--nii',
            default='../raw/nii/{2}/{0}-{1:04d}-{2}-{3}.nii',
            help='input file;' + hp.nii)

    parser.add_argument('--irritation',
            default='../data/irr/{2}/{0}-{1:04d}-{2}-{3}.irr',
            help='input file;' + hp.irritation)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('--session',
            default='../data/ses/{2}/{0}-{1:04d}-{2}-{3}.ses',
            help='output file;' + hp.session)

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-nii2ses.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for the setup of a session instance
########################################################################

    parser.add_argument('--epi-code',
            type=int,
            help=hp.epi_code)

    foreground_handling = parser.add_mutually_exclusive_group()

    foreground_handling.add_argument('--detect-foreground',
            action='store_true',
            help=hp.detect_foreground)

    foreground_handling.add_argument('--set-foreground',
            action='store_true',
            help=hp.set_foreground)

    parser.add_argument('--foreground',
            default='../raw/frg/{2}/{0}-{1:04d}-{2}-{3}-foreground.nii.gz',
            help=hp.set_foreground)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('--protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--datetime',
            help=hp.datetime)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Miscellaneous
########################################################################

    lock_handling = parser.add_mutually_exclusive_group()

    lock_handling.add_argument('--remove-lock',
            action='store_true',
            help=hp.remove_lock.format('population space'))

    lock_handling.add_argument('--ignore-lock',
            action='store_true',
            help=hp.ignore_lock.format('population space'))

    file_handling = parser.add_mutually_exclusive_group()

    file_handling.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('population space'))

    file_handling.add_argument('-s', '--skip',
            action='store_true',
            help=hp.skip.format('population space'))

    parser.add_argument('-v', '--verbose',
            action='count',
            default=0,
            help=hp.verbose)

########################################################################
# Multiprocessing
########################################################################

    parser.add_argument('-j', '--cores',
            type=int,
            help=hp.cores)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...lock import Lock

from ...load import load_block_irritation, load_session

from ...name import Identifier

from ...protocol import layout_dummy

from ...irritation import Block

from ...session import Session, fmrisetup

from ...nifti import nii2session

import nibabel as ni

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args, fall_back=args.irritation)

    if df is None:
        sys.exit()

    if not 'epi' in df.columns:
        print("""
No column epi found. You need to provide an EPI code via --epi-code, and
it must be integer, within [-3,3], and not null.""")

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_dummy(df_layout, 'nii',
            template=args.nii,
            strftime=args.strftime
            )

    layout_dummy(df_layout, 'foreground',
            template=args.foreground,
            strftime=args.strftime
            )

    layout_dummy(df_layout, 'irr',
            template=args.irritation,
            strftime=args.strftime
            )

    layout_dummy(df_layout, 'file',
            template=args.session,
            strftime=args.strftime
            )

    ####################################################################
    # Apply wrapper
    ####################################################################

    df['locked'] = False

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        try:
            dfile = os.path.dirname(r.file)
            if dfile and not isdir(dfile):
                os.makedirs(dfile)
        except Exception as e:
            print('{}: {}'.format(name.name(), e))

        wrapper(name              = name,
                df                = df,
                index             = r.Index,
                remove_lock       = args.remove_lock,
                ignore_lock       = args.ignore_lock,
                force             = args.force,
                skip              = args.skip,
                verbose           = args.verbose,
                file              = r.file,

                file_nii = r.nii,
                file_irr = r.irr,

                epi_code = r.epi,
                detect_foreground = args.detect_foreground,
                set_foreground = args.set_foreground,
                foreground = r.foreground,
                )

    it =  df_layout.itertuples()

    if len(df_layout) > 1 and ((args.cores is None) or (args.cores > 1)):
        try:
            pool = ThreadPool(args.cores)
            results = pool.map(wm, it)
            pool.close()
            pool.join()
        except Exception as e:
            pool.close()
            pool.terminate()
            print('Pool execution has been terminated')
            print(e)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']
    else:
        try:
            print('Process protocol entries sequentially')
            for r in it:
                wm(r)
        finally:
            files = df_layout.ix[df.locked, 'file'].values
            if len(files) > 0:
                for f in files:
                    print('Unlock: {}'.format(f))
                    os.remove(f)
            del df['locked']

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

########################################################################

def wrapper(name, df, index, remove_lock, ignore_lock, force, skip,
        verbose, file, file_nii , file_irr, epi_code,
        detect_foreground, set_foreground, foreground):

    if isfile(file):
        instance = load_session(file, name, df, index, verbose)
        if type(instance) is Lock:
            if remove_lock or ignore_lock:
                if verbose:
                    print('{}: Remove lock'.format(name.name()))
                instance.unlock()
                if remove_lock:
                    return
            else:
                if verbose:
                    print('{}: Locked'.format(name.name()))
                return
        else:
            if df.ix[index,'valid'] and not force:
                if verbose:
                    print('{}: Valid'.format(name.name()))
                return
            else:
                if skip:
                    if verbose:
                        print('{}: Invalid'.format(name.name()))
                    return

    if verbose:
        print('{}: Lock: {}'.format(name.name(), file))

    lock = Lock(name, 'fmrifit', file)
    df.ix[index, 'locked'] = True
    lock.save(file)
    df.ix[index,'valid'] = True

    ####################################################################
    # Load irritation instance from disk
    ####################################################################

    irritation = load_block_irritation(file_irr, name, df, index, verbose)
    if lock.conditional_unlock(df, index, verbose):
        return

    ################################################################
    # Load image data
    ################################################################

    try:
        img = ni.load(file_nii)
        if verbose:
            print('{}: Read: {}'.format(name.name(), file_nii))
    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to read: {}'.format(name.name(), file_nii))
        print('{}: Exception: {}'.format(name.name(), e))

    if lock.conditional_unlock(df, index, verbose):
        return

    ####################################################################
    # Create session instance
    ####################################################################

    try:
        session = nii2session(
                name=name,
                nii=img,
                epi_code=epi_code)

        fmrisetup(session = session, irritation = irritation)

        if detect_foreground:
            if verbose:
                print('{}: Detect foreground'.format(name.name()))
            session.fit_foreground()
        elif set_foreground:
            foreground = ni.load(foreground)
            if verbose:
                print('{}: Set foreground: {}'.format(name.name(), foreground))
            session.set_foreground(foreground.data)

        if verbose:
            print('{}: Save: {}'.format(name.name(), file))

        session.save(file)
        df.ix[index,'locked'] = False

    except Exception as e:
        df.ix[index,'valid'] = False
        print('{}: Unable to create: {}'.format(name.name(), file))
        print('{}: Exception: {}'.format(name.name(), e))
        lock.conditional_unlock(df, index, verbose, True)
        return



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='fmristats',
    version='0.0.2',
    description='Rigorous statistical estimation of FMRI models',
    long_description=long_description,
    url='https://fmristats.github.io/',
    author='Thomas W. D. Möbius',
    author_email='moebius@medinfo.uni-kiel.de',
    license='GPLv3+',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='fmri mri statistics meta-analysis meta-regression imaging neuroimaging',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['numpy', 'scipy', 'scikit-image', 'pandas',
        'statsmodels', 'matplotlib', 'seaborn', 'nibabel'],
    entry_points={
        'console_scripts': [
            # general api
            'fmriassessment  = fmristats.cmd.api.fmriassessment:cmd',
            'fmriati         = fmristats.cmd.api.fmriati:cmd',
            'fmrifit         = fmristats.cmd.api.fmrifit:cmd',
            'fmriirritation  = fmristats.cmd.api.fmriirritation:cmd',
            'fmririgid       = fmristats.cmd.api.fmririgid:cmd',
            'fmrisample      = fmristats.cmd.api.fmrisample:cmd',
            'fmripop         = fmristats.cmd.api.fmripop:cmd',
            'fmriprune       = fmristats.cmd.api.fmriprune:cmd',

            # wrapper
            'fsl4prune = fmristats.cmd.wrp.fsl4prune:cmd',
            'fsl4pop   = fmristats.cmd.wrp.fsl4pop:cmd',

            # import from third party formats
            'mat2irr = fmristats.cmd.imp.mat2irr:cmd',
            'nii2img = fmristats.cmd.imp.nii2img:cmd',
            'nii2ses = fmristats.cmd.imp.nii2ses:cmd',
            'csv2protocol = fmristats.cmd.imp.csv2protocol:cmd_pkl',
            'csv2covariate = fmristats.cmd.imp.csv2protocol:cmd_cov',

            # plot
            'fit2plot = fmristats.cmd.plt.fit2plot:cmd',
            'ref2plot = fmristats.cmd.plt.ref2plot:cmd',
            #'ses2plot = fmristats.cmd.plt.ses2plot:cmd',

            # query
            'fmrimap         = fmristats.cmd.qry.fmrimap:cmd',
            'fmriinfo        = fmristats.cmd.qry.fmriinfo:cmd',
            'fmriprotocol    = fmristats.cmd.qry.fmriprotocol:cmd',
            'fmricovariate   = fmristats.cmd.qry.fmricovariate:cmd',

            # export to third party formats
            'fit2nii = fmristats.cmd.exp.fit2nii:cmd',

        ],
    },
)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Quality assessment statistics for image quality, ability to track head
movements, and descriptive statistics of fitted effect fields.

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('--session',
            default='../data/ses/{2}/{0}-{1:04d}-{2}-{3}.ses',
            help='input file;' + hp.session)

    parser.add_argument('--reference-maps',
            default='../data/ref/{2}/{0}-{1:04d}-{2}-{3}.ref',
            help='input file;' + hp.reference_maps)

    parser.add_argument('--fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='input file;' + hp.sfit)

    parser.add_argument('--population-space',
            default='resolution',
            help=hp.population_space)

    parser.add_argument('--scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help=hp.scale_type)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    parser.add_argument('--protocol',
            help=hp.protocol)

    parser.add_argument('--cohort',
            help=hp.cohort)

    parser.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    parser.add_argument('--datetime',
            help=hp.datetime)

    parser.add_argument('--paradigm',
            help=hp.paradigm)

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fitassessment.pkl',
            help=hp.protocol_log)

    parser.add_argument('--assess-sessions',
            action='store_true',
            help="""whether to assess descriptive statistics of the
            sessions""")

    parser.add_argument('--assess-reference-maps',
            action='store_true',
            help="""whether to assess descriptive statistics of the
            reference maps""")

    parser.add_argument('--assess-fits',
            action='store_true',
            help="""whether to assess descriptive statistics of the
            fits""")

########################################################################
# Miscellaneous
########################################################################

    parser.add_argument('-v', '--verbose',
            action='store_true',
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...load import load_result, load_session, load_refmaps

from ...name import Identifier

from ...protocol import layout_dummy, layout_sdummy

#from ... import Session, ReferenceMaps, PopulationMap, SignalModel, Result

import pandas as pd

from pandas import Series

import datetime

import sys

import os

from os.path import isfile, isdir, join

import numpy as np

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_dummy(df_layout, 'ses',
            template=args.session,
            strftime=args.strftime
            )

    layout_dummy(df_layout, 'ref',
            template=args.reference_maps,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'fit',
            template=args.fit,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    ####################################################################
    # Add fields for summary statistics
    ####################################################################

    if args.assess_sessions:
        df['muf'] = Series(np.nan, index=df.index, dtype=float)
        df['mub'] = Series(np.nan, index=df.index, dtype=float)
        df['mfb'] = Series(np.nan, index=df.index, dtype=float)
        df['sdf'] = Series(np.nan, index=df.index, dtype=float)
        df['sdb'] = Series(np.nan, index=df.index, dtype=float)
        df['cvf'] = Series(np.nan, index=df.index, dtype=float)
        df['cvb'] = Series(np.nan, index=df.index, dtype=float)

    if args.assess_reference_maps:
        df['cyc']= Series(0, index=df.index, dtype=int)
        df['scn']= Series(0, index=df.index, dtype=int)
        df['cyo']= Series(0, index=df, dtype=int)
        df['sco']= Series(0, index=df, dtype=int)

    if args.assess_fits:
        df['vol'] = Series(np.nan, index=df, dtype=float)
        df['occ'] = Series(0, index=df, dtype=int)

    #################################################################
    # Apply wrappers
    #################################################################

    if args.assess_sessions:
        for r in df_layout.itertuples():
            name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)
            wrapper_session(
                    name     = name,
                    df       = df,
                    index    = r.Index,
                    filename = r.ses,
                    verbose  = args.verbose
                    )

    if args.assess_reference_maps:
        for r in df_layout.itertuples():
            name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)
            wrapper_refmaps(
                    name     = name,
                    df       = df,
                    index    = r.Index,
                    filename = r.ref,
                    verbose  = args.verbose
                    )

    if args.assess_fits:
        for r in df_layout.itertuples():
            name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)
            wrapper_fits(
                    name     = name,
                    df       = df,
                    index    = r.Index,
                    filename = r.fit,
                    urname   = args.population_space,
                    verbose  = args.verbose
                    )

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

#######################################################################

def wrapper_session(name, df, index, filename, verbose):
    session = load_session(filename, name, df, index, verbose)
    if df.ix[index,'valid'] == False:
        return

    dstats = session.descriptive_statistics()
    dstats_names = ['muf', 'mub', 'mfb', 'sdf', 'sdb', 'cvf', 'cvb']
    dstats = Series(dstats, index=dstats_names)
    for s in dstats_names:
        df.ix[index,s] = dstats[s]

def wrapper_refmaps(name, df, index, filename, verbose):
    reference_maps = load_refmaps(filename, name, df, index, verbose)
    if df.ix[index,'valid'] == False:
        return

    dstats = reference_maps.descriptive_statistics()
    dstats_names = ['cyc', 'scn', 'cyo', 'sco']
    dstats = Series(dstats, index=dstats_names)
    for s in dstats_names:
        df.ix[index,s] = dstats[s]

def wrapper_fits(name, df, index, filename, urname, verbose):
    result = load_result(filename, name, df, index, urname, verbose)
    if df.ix[index,'valid'] == False:
        return

    dstats = result.descriptive_statistics()
    dstats_names = ['vol', 'occ']
    dstats = Series(dstats, index=dstats_names)
    for s in dstats_names:
        df.ix[index,s] = dstats[s]



"""
Will parse a the output of ... to a irritation.
"""

########################################################################
#
# Tool
#
########################################################################

from .name import Identifier

from .irritation import Block

def mat2irritation(mat, name):
    """
    Will parse a the output of ... to a block irritation.

    Parameters
    ----------
    mat : dict
        dict returned by scipy.io.loadmat(args.input)
    name : Identifier
        identifier of the sesson.
    """
    assert type(name) is Identifier, 'name must be of type Identifier'
    block_number = mat['names'].shape[-1]
    names = [mat['names'].ravel()[i][0] for i in range(block_number)]
    onsets = { names[i] : mat['onsets'].ravel()[i].ravel() for i in range(block_number)}
    durations = { names[i] : mat['durations'].ravel()[i].ravel()[0] for i in range(block_number)}
    return Block(name=name, names=names, onsets=onsets, durations=durations)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Handling observations

"""

import numpy as np

def create_slice_time_matrix(shape, epi_code, temporal_resolution,
        interleaved=None):
    """
    Creates a time vector from parameters

    Parameters
    ----------
    shape : tuple, type: int
    epi_code : int or None
    temporal_resolution : float
        The temporal resolution.
    interleaved : bool

    Returns
    -------
    slice_time : ndarray of float, shape (M,) or (M,N,O,P)
        The time vector that holds the slices time information. The
        vector holds the start point of the time at which a particular
        slice has been measured.  If ep is None, slice_time has
        shape (M,) with M equal to the number of full scan cycles.  If
        ep is int, then slice_time has
        the same shape as the data.

    """
    # TODO: implement interleaved
    assert not interleaved, 'sorry, interleaved not implemented yet'
    n = shape[0]
    m = shape[abs(epi_code)]
    slice_time = np.arange(
            start = 0,
            stop  = n*temporal_resolution,
            step  = temporal_resolution/m).reshape((n,m))
    return slice_time



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Converts a csv to protocol

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

    parser.add_argument('csv', help="""csv file""")

    parser.add_argument('output', help="""output file""")

    parser.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

    return parser

def cmd_pkl():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args, False)

def cmd_cov():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args, True)

cmd_pkl.__doc__ = __doc__
cmd_cov.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

import pandas as pd

########################################################################

def call(args, drop):
    df = pd.read_csv(args.csv,
            dtype={
                'cohort':str,
                'id':int,
                'paradigm':str,
                'date':str,
                'epi':int,
                'valid':bool})

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    if not hasattr(df, 'valid'):
        df['valid'] = True

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df.date, format=args.strftime)
        df.set_index(
                keys=['cohort', 'id', 'paradigm', 'date'],
                drop=drop,
                inplace=True,
                verify_integrity=True)
        df.sort_index(inplace=True)
        if not drop:
            df.cohort = df.cohort.astype('category')
            df.paradigm = df.paradigm.astype('category')

    elif 'paradigm' in df.columns:
        df.set_index(
                keys=['cohort', 'id', 'paradigm'],
                drop=drop,
                inplace=True,
                verify_integrity=True)
        df.sort_index(inplace=True)
        if not drop:
            df.cohort = df.cohort.astype('category')
            df.paradigm = df.paradigm.astype('category')

    else:
        df.set_index(
                keys=['cohort', 'id'],
                drop=drop,
                inplace=True,
                verify_integrity=True)
        df.sort_index(inplace=True)
        if not drop:
            df.cohort = df.cohort.astype('category')

    df.to_pickle(args.output)



# Copyright 2016-2017 Thomas W. D. Möbius
#
# This file is part of fmristats.
#
# fmristats is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# fmristats is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# It is not allowed to remove this copy right statement.

"""

Export fit to Nifti1

"""

########################################################################
#
# Command line program
#
########################################################################

import fmristats.cmd.hp as hp

import argparse

def create_argument_parser():
    parser = argparse.ArgumentParser(
            description=__doc__,
            epilog=hp.epilog)

########################################################################
# Input arguments
########################################################################

    parser.add_argument('--fit',
            default='../data/fit/{2}/{4}/{5}/{0}-{1:04d}-{2}-{3}-{4}-{5}.fit',
            help='input file;' + hp.sfit)

########################################################################
# Output arguments
########################################################################

    parser.add_argument('--nii',
            default='../results/nii/{2}/{0}-{1:04d}-{2}-{3}-{4}-{5}-{{}}-{{}}.nii.gz',
            help="""output file""")

    parser.add_argument('-o', '--protocol-log',
            default='logs/{}-fit2nii.pkl',
            help=hp.protocol_log)

########################################################################
# Arguments specific for the RSM Signal Model: where to
########################################################################

    where_to_fit = parser.add_argument_group(
            """where to""",
            """By default, fit2nii will respect the brain mask which is
            saved in the population map. This default behaviour can, of
            course, be changed.""")

    where_to_fit.add_argument('--ignore-mask',
            action='store_true',
            help="""Ignore any brain masks saved in the respective
            population map.""")

    # TODO: mask: not implemented yet
    #where_to_fit.add_argument('--mask',
    #        action='store_true',
    #        help=hp.population_mask)

########################################################################
# Arguments specific for using the protocol API
########################################################################

    to_process = parser.add_argument_group(
            """specifying the protocol entries to process""",
            """Arguments which give control which protocol entries to
            process. If no protocol file is given, it will be checked
            if files being processed comply to the given information.""")

    to_process.add_argument('--protocol',
            help=hp.protocol)

    to_process.add_argument('--cohort',
            help=hp.cohort)

    to_process.add_argument('--id',
            type=int,
            nargs='+',
            help=hp.j)

    to_process.add_argument('--datetime',
            help=hp.datetime)

    to_process.add_argument('--paradigm',
            help=hp.paradigm)

    to_process.add_argument('--strftime',
            default='%Y-%m-%d-%H%M',
            help=hp.strftime)

    which_to_process = parser.add_argument_group(
            """specifying fit and population space""",
            """If in-  and output files are specified as templates, set
            some of the fixed terms in the template.""")

    which_to_process.add_argument('--scale-type',
            default='max',
            choices=['diagonal','max','min'],
            help=hp.scale_type)

    which_to_process.add_argument('--population-space',
            default='reference',
            help=hp.population_space)

########################################################################
# Arguments which field to extract
########################################################################

    parser.add_argument('--parameter',
            type=str,
            nargs='+',
            default=['intercept', 'activation'],
            help="""which parameter field or which parameter fields to
            plot.  Available options are `intercept`, `activation`, or
            `time`. If you want to plot more than one parameter field,
            then separate these by a space.""")

    parser.add_argument('--value',
            type=str,
            nargs='+',
            default=['point', 'tstatistic'],
            #default=['point', 'stderr', 'tstatistic'],
            help="""which summary statistic to choose for the parameters
            in PARAMETER. Available options are `point` (the point
            estimate), `stderr` (standard deviation of the point
            estimator), or `tstatistic` (the t-test statistic testing
            whether the respected parameter is non zero). If you want to
            plot more than one parameter field, then separate these by a
            space.""")

########################################################################
# Arguments specific for the application of masks
########################################################################

    parser.add_argument('--round-to-integer',
            action='store_true',
            help="""if this flag is set and the output field is the
            intercept of a fit, then round the grey scales of the
            intercept field to the next nearest integer, and cast the
            array to integer data type.""")

########################################################################
# Miscellaneous
########################################################################

    parser.add_argument('-f', '--force',
            action='store_true',
            help=hp.force.format('result'))

    parser.add_argument('-v', '--verbose',
            action='store_true',
            help=hp.verbose)

    return parser

def cmd():
    parser = create_argument_parser()
    args = parser.parse_args()
    call(args)

cmd.__doc__ = __doc__

########################################################################
#
# Load libraries
#
########################################################################

from ..df import get_df

from ...lock import Lock

from ...load import load_result

from ...name import Identifier

from ...protocol import layout_dummy, layout_sdummy

from ...smodel import Result

from ...plot import picture

from ...nifti import image2nii

import pandas as pd

import datetime

import sys

import os

from os.path import isfile, isdir, join

from multiprocessing.dummy import Pool as ThreadPool

import numpy as np

import nibabel as ni

########################################################################

def call(args):
    output = args.protocol_log.format(
            datetime.datetime.now().strftime('%Y-%m-%d-%H%M'))

    if args.strftime == 'short':
        args.strftime = '%Y-%m-%d'

    ####################################################################
    # Parse protocol
    ####################################################################

    df = get_df(args, fall_back=args.fit)

    if df is None:
        sys.exit()

    ####################################################################
    # Add file layout
    ####################################################################

    df_layout = df.copy()

    layout_sdummy(df_layout, 'nii',
            template=args.nii,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    layout_sdummy(df_layout, 'file',
            template=args.fit,
            urname=args.population_space,
            scale_type=args.scale_type,
            strftime=args.strftime
            )

    ####################################################################
    # Respect the mask mask
    ####################################################################

    mask = True

    if args.ignore_mask:
        mask = False

    ####################################################################
    # Apply wrapper
    ####################################################################

    def wm(r):
        name = Identifier(cohort=r.cohort, j=r.id, datetime=r.date, paradigm=r.paradigm)

        wrapper(name                  = name,
                df                    = df,
                index                 = r.Index,
                filename              = r.file,
                template              = r.nii,
                verbose               = args.verbose,
                vb                    = args.population_space,
                params                = args.parameter,
                values                = args.value,
                mask                  = mask,
                round_to_integer      = args.round_to_integer)

    it =  df_layout.itertuples()

    for r in it:
        wm(r)

    ####################################################################
    # Write protocol
    ####################################################################

    if args.verbose:
        print('Save: {}'.format(output))

    dfile = os.path.dirname(output)
    if dfile and not isdir(dfile):
       os.makedirs(dfile)

    df.to_pickle(output)

########################################################################

def wrapper(name, df, index, filename, template, verbose,
        vb, params, values, mask, round_to_integer):

    ####################################################################
    # Load fit from disk
    ####################################################################

    result = load_result(filename, name, df, index, vb, verbose)
    if df.ix[index,'valid'] == False:
        return

    if verbose > 1:
        print('{}: Description of the fit:'.format(result.name.name()))
        print(result.describe())

    result.mask(mask, verbose)

    for param in params:
        if param in result.parameter_dict.keys():
            for value in values:

                fname = template.format(param, value)
                dfile = os.path.dirname(fname)
                if dfile and not isdir(dfile):
                   os.makedirs(dfile)
                field = result.get_field(param=param, value=value)

                if value == 'intercept' and round_to_integer:
                    field = field.round()

                if verbose:
                    print('{}: Save ({}) to {}'.format(
                        name.name(), field.name, fname))
                ni.save(image2nii(field), fname)
